"""
Plot for the probability of contact formation with fit
for the final paper
"""
from collections import namedtuple

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from scipy import stats, special

from utils.misc import block_avg
from utils.paths import CM_PATH, ANALYSIS_PATH

######################## functions
def logistic(x, l, k, x0):
    '''Logistic function'''
    return l / (1 + np.exp(-k*(x - x0)))

def tstar_k_draw(ax, param):
    # dran dashed line
    l = param[0]
    k = param[1]
    ts = param[2]
    ax.vlines(
        x=ts,
        ymin=0, ymax=logistic(ts, l, k, ts),
        linestyle="dashed",
        color="black")
    # tstar label
    ax.text(
        x=ts, y=-0.125,
        s='$t^*$')
    # k line
    derivative = l*k/4
    distance_ts = 0.25e4
    intercept = logistic(ts, l, k, ts) - derivative*ts
    # ax.plot(
    #     [ts-distance_ts, ts+distance_ts],
    #     [derivative*(ts-distance_ts) + intercept, derivative*(ts+distance_ts) + intercept],
    #     linewidth=1.5,
    #     color='green'
    # )
    # slope triangle
    dts = 0.075e4
    ax.fill(
        [ts - dts, ts + dts, ts + dts],
        [derivative*(ts - dts) + intercept, derivative*(ts - dts) + intercept, derivative*(ts + dts) + intercept],
        facecolor='#f1faee',
        edgecolor='#d4a373',
        alpha=.7,
        linewidth=1.5,
        zorder=10)
    # k label
    ax.text(
        x=ts+dts*2, y=(derivative*ts + intercept)-0.035,
        s='$Lk/4$', fontsize=20)

def chisquare(f_obs, f_exp=None, ddof=0, axis=0):
    return power_divergence(f_obs, f_exp=f_exp, ddof=ddof, axis=axis,
                            lambda_="pearson")

# Map from names to lambda_ values used in power_divergence().
_power_div_lambda_names = {
    "pearson": 1,
    "log-likelihood": 0,
    "freeman-tukey": -0.5,
    "mod-log-likelihood": -1,
    "neyman": -2,
    "cressie-read": 2/3
}

def _count(a, axis=None):
    if hasattr(a, 'count'):
        num = a.count(axis=axis)
        if isinstance(num, np.ndarray) and num.ndim == 0:
            # In some cases, the `count` method returns a scalar array (e.g.
            # np.array(3)), but we want a plain integer.
            num = int(num)
    else:
        if axis is None:
            num = a.size
        else:
            num = a.shape[axis]
    return num

Power_divergenceResult = namedtuple('Power_divergenceResult',
                                    ('statistic', 'pvalue'))

def power_divergence(f_obs, f_exp=None, ddof=0, axis=0, lambda_=None):
    # Convert the input argument `lambda_` to a numerical value.
    if isinstance(lambda_, str):
        if lambda_ not in _power_div_lambda_names:
            names = repr(list(_power_div_lambda_names.keys()))[1:-1]
            raise ValueError("invalid string for lambda_: {0!r}.  Valid strings "
                             "are {1}".format(lambda_, names))
        lambda_ = _power_div_lambda_names[lambda_]
    elif lambda_ is None:
        lambda_ = 1

    f_obs = np.asanyarray(f_obs)

    if f_exp is not None:
        f_exp = np.asanyarray(f_exp)
    else:
        # Ignore 'invalid' errors so the edge case of a data set with length 0
        # is handled without spurious warnings.
        with np.errstate(invalid='ignore'):
            f_exp = f_obs.mean(axis=axis, keepdims=True)

    # `terms` is the array of terms that are summed along `axis` to create
    # the test statistic.  We use some specialized code for a few special
    # cases of lambda_.
    if lambda_ == 1:
        # Pearson's chi-squared statistic
        terms = (f_obs.astype(np.float64) - f_exp)**2 / f_exp
    elif lambda_ == 0:
        # Log-likelihood ratio (i.e. G-test)
        terms = 2.0 * special.xlogy(f_obs, f_obs / f_exp)
    elif lambda_ == -1:
        # Modified log-likelihood ratio
        terms = 2.0 * special.xlogy(f_exp, f_exp / f_obs)
    else:
        # General Cressie-Read power divergence.
        terms = f_obs * ((f_obs / f_exp)**lambda_ - 1)
        terms /= 0.5 * lambda_ * (lambda_ + 1)

    stat = terms.sum(axis=axis)

    num_obs = _count(terms, axis=axis)
    ddof = np.asarray(ddof)
    p = stats.distributions.chi2.sf(stat, num_obs - 1 - ddof)

    return Power_divergenceResult(stat, p)

######################## main
matplotlib.rcParams['font.size'] = 24
matplotlib.rcParams['text.usetex'] = True
matplotlib.rcParams['font.family'] = 'modern computer'

protein = ['1ucs', '1ucs']
datasets = ['29Oct22_CV_2Nov22_CV', '29Oct22_CV_2Nov22_CV']

# Contact map
cm_native = [
    np.load(CM_PATH / f'{protein[0]}_cm.npy'),
    np.load(CM_PATH / f'{protein[1]}_cm.npy')
]

# Number of native contacts
total_contacts = [
    (np.triu(cm_native[0]) != 0).sum(),
    (np.triu(cm_native[1]) != 0).sum()
]

# load GE for the native contacts, together with all the 
# residue pairs involved in a loop (hence in a contact)
GE_native = [
    np.load(CM_PATH / f'GE_{protein[0]}_native.npy', allow_pickle=True),
    np.load(CM_PATH / f'GE_{protein[1]}_native.npy', allow_pickle=True)
]

contact_avg = [
    np.load(ANALYSIS_PATH / protein[0] / f'contact_avg_{datasets[0]}.npy' ),
    np.load(ANALYSIS_PATH / protein[1] / f'contact_avg_{datasets[1]}.npy' )
]
parameters = [
    np.load(ANALYSIS_PATH / protein[0] / f'logistic_fit_para_{datasets[0]}.npy' ),
    np.load(ANALYSIS_PATH / protein[1] / f'logistic_fit_para_{datasets[1]}.npy' )
]

idx = [38, 105]

# Plot
fig, axs = plt.subplots(
    2, figsize=(11, 10),
    gridspec_kw={"hspace":0.25})

# axs[0].set_title(
#     "$L / \\left[ 1 + \\exp \\left(- k (t - t^*)\\right) \\right]$",
#     pad=25)

for ax in axs.flatten():
    ax.set_ylim(0,1)
    ax.grid(alpha=.4)
    ax.spines.top.set_visible(False)
    ax.spines.right.set_visible(False)
    ax.ticklabel_format(axis="x", scilimits=(0,4))
    ax.yaxis.set_minor_locator(plt.MultipleLocator(0.1))
    ax.yaxis.set_major_locator(plt.MultipleLocator(0.2))
    ax.xaxis.set_minor_locator(plt.MultipleLocator(0.1e4))
    ax.xaxis.set_major_locator(plt.MultipleLocator(0.5e4))

# 1ucs
res1 = GE_native[0][idx[0]][0][0]
res2 = GE_native[0][idx[0]][0][1]
print(res1+1, res2+1)

timeseries = contact_avg[0][:, res1, res2]
t_block, _ = block_avg(timeseries, 100)
lensim = len(timeseries)

axs[0].plot(timeseries, alpha=.5, c="#457b9d")
x_avg = np.linspace(0, lensim, lensim//100)
axs[0].plot(x_avg, t_block, 'o', c='#1d3557', alpha=.3)
logis = logistic(x_avg, *parameters[0][idx[0]])
axs[0].plot(x_avg, logis, c='#e63946', linewidth=2)\

tstar_k_draw(axs[0], parameters[0][idx[0]])

axs[0].set_xlim(-0.1e4, 2.60e4)

axs[0].set_ylabel(r"$\langle \sigma_{6, 25} \rangle (t)$", labelpad=15)

# goodness
# ddof = len(t_block) - 3
# chi, pv = chisquare(f_obs=t_block, f_exp=logis, ddof=ddof)
# axs[0].text(
#     x=.5, y=.1,
#     s=f'$\chi^2_o$ reduced: {chi/ddof:.3f}; $P(\chi^2 > \chi^2_o)$: {pv:.2f}',
#     fontsize=15, transform=axs[0].transAxes)

# 1srl
res1 = GE_native[1][idx[1]][0][0]
res2 = GE_native[1][idx[1]][0][1]
print(res1+1, res2+1)

timeseries = contact_avg[1][:, res1, res2]
t_block, _ = block_avg(timeseries, 100)
lensim = len(timeseries)

l1,  = axs[1].plot(timeseries, alpha=.5, c="#457b9d")
x_avg = np.linspace(0, lensim, lensim//100)
l2,  = axs[1].plot(x_avg, t_block, 'o', c='#1d3557', alpha=.3)
logis = logistic(x_avg, *parameters[1][idx[1]])
l3,  = axs[1].plot(x_avg, logis, c='#e63946', linewidth=2)

tstar_k_draw(axs[1], parameters[1][idx[1]])

axs[1].set_xlim(-0.1e4, 2.60e4)

axs[1].set_ylabel(r"$\langle \sigma_{23, 37} \rangle (t)$", labelpad=15)
axs[1].set_xlabel("Time [MD step]", labelpad=15)

# goodness
# ddof = len(t_block) - 3
# chi, pv = chisquare(f_obs=t_block, f_exp=logis, ddof=ddof)
# axs[1].text(
#     x=.5, y=.1,
#     s=f'$\chi^2_o$ reduced: {chi/ddof:.3f}; $P(\chi^2 > \chi^2_o)$: {pv:.2f}',
#     fontsize=15, transform=axs[1].transAxes)

# legend
axs[1].legend(
    handles = [l1,l2,l3],
    labels=['Timeseries', 'Block average', 'Fit $L / \\left[ 1 + \\exp \\left(- k (t - t^*)\\right) \\right]$'],
    loc='upper center', bbox_to_anchor=(0.45, 2.55),
    ncol=3, fancybox=True, shadow=True,
    fontsize=20)

fig.savefig(ANALYSIS_PATH / "1ucs" / "img" / f'contact_probability_fit_1ucs_contacts_{idx[0]}_1ucs_contact_{idx[1]}.pdf', bbox_inches='tight')
