"""
Compute, save and plot the contact map probabilities for the IE_- and IT
ensembles of 1ucs. The range selection is user-specified.
"""
import os
from pathlib import Path

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import MDAnalysis as mda
from MDAnalysis.analysis import distances
import seaborn as sns

from utils.paths import ANALYSIS_PATH, CM_PATH, DATA_PATH
from utils.misc import timeseries_mask
from visualization.viztool import textBox

###########################################################
prot_name = '1ucs'
dataset_list = ['29Oct22_CV', '2Nov22_CV']

# Q parameters
gamma = 1.2
# Second RC parameters
rc_name = "wGEa_threshold_0.50_exponent_4.00"
r = 1500

# Ensemble range. NB these do not have to overlap
## Based on the wGEa_0.5_3 plot

# points to define IT: rectangle
QlimIT = (0.5, 0.75)
GElimIT_high = (0.1, 0.05)
GElimIT_low = (-0.1, -0.15)

# Points to define IE-: parallelepiped
QlimIE = (0.32, 0.46)
GElimIE_high = (-0.45, -0.6)
GElimIE_low = (-0.6, -0.75)
###########################################################
# def a linear function to describe parallelepiped in cartesian space
def linear_ie_high(q):
    m = (GElimIE_high[0] - GElimIE_high[1])/(QlimIE[0] - QlimIE[1])
    interc = (-GElimIE_high[0]*QlimIE[1] + GElimIE_high[1]*QlimIE[0])/(QlimIE[0] - QlimIE[1])
    # print(f"IE upper: m = {m}; q = {interc}")
    return q*m + interc
def linear_ie_low(q):
    m = (GElimIE_low[0] - GElimIE_low[1])/(QlimIE[0] - QlimIE[1])
    interc = (-GElimIE_low[0]*QlimIE[1] + GElimIE_low[1]*QlimIE[0])/(QlimIE[0] - QlimIE[1])
    # print(f"IE lower: m = {m}; q = {interc}")
    return q*m + interc
def linear_it_high(q):
    m = (GElimIT_high[0] - GElimIT_high[1])/(QlimIT[0] - QlimIT[1])
    interc = (-GElimIT_high[0]*QlimIT[1] + GElimIT_high[1]*QlimIT[0])/(QlimIT[0] - QlimIT[1])
    # print(f"IT upperr: m = {m}; q = {interc}")
    return q*m + interc
def linear_it_low(q):
    m = (GElimIT_low[0] - GElimIT_low[1])/(QlimIT[0] - QlimIT[1])
    interc = (-GElimIT_low[0]*QlimIT[1] + GElimIT_low[1]*QlimIT[0])/(QlimIT[0] - QlimIT[1])
    # print(f"IT lower: m = {m}; q = {interc}")
    return q*m + interc

PROT_PATH = ANALYSIS_PATH / prot_name
dataset = ''
for d in dataset_list:
    dataset += f'_{d}'
dataset = dataset[1:]

# graphical settings
matplotlib.rcParams['font.size'] = 24
matplotlib.rcParams['text.usetex'] = True
matplotlib.rcParams['font.family'] = 'modern computer'

# load Q and G' data
Q = np.load(PROT_PATH / f'Q_{dataset}.npy', allow_pickle=True)
GE = np.load(PROT_PATH / f'{rc_name}_r{r}_{dataset}.npy', allow_pickle=True)
id_map = np.loadtxt(PROT_PATH / f"ID_mapping_{dataset}.csv", delimiter=",", dtype=int)

# simulations path
simulations = [str(DATA_PATH / date / sim)
               for date in dataset_list
               for sim in os.listdir(DATA_PATH / date)
               if sim != 'logs']

# Universes for all simulations
universes = []
for sim in simulations:
    simdir = Path(sim)
    PSF = simdir / f'{prot_name}_init.data'
    eqDCD = simdir / f'{prot_name}.dcd'
    universes.append(mda.Universe(str(PSF), str(eqDCD)))
universes = np.array(universes)

# Checks and reloading of simulations and universes if Q and RMSD has been loaded
# from disk
assert Q.shape[0] == id_map.shape[0]
# this means that the files has been loaded
# Hence, simulations and universes need to be ordered
# in the same manner
assert len(simulations) == id_map.shape[0]
for sim, id_tuple in zip(simulations, id_map):
    id_from_sim = int(sim.split("_")[-1])
    if id_tuple[1] != id_from_sim:
        print(f"Subsituing {id_from_sim} with {id_tuple[1]} in position {id_tuple[0]}")

        str_split = simulations[id_tuple[0]].split("_")
        str_split[-1] = str(id_tuple[1])
        simulations[id_tuple[0]] = "".join(str_split)

        simdir = Path(simulations[id_tuple[0]])
        PSF = simdir / f'{prot_name}_init.data'
        eqDCD = simdir / f'{prot_name}.dcd'
        universes[id_tuple[0]] = mda.Universe(str(PSF), str(eqDCD))

# subsampling Q in order to hav the same points with G'
mask = timeseries_mask(r, Q[0])
mask_fancy = np.flatnonzero(mask)
Q_sub = [q[mask_fancy] for q in Q]

cm_native = np.load(CM_PATH / f'{prot_name}_cm.npy')
cm_native = np.triu(cm_native)
cm_mod = np.copy(cm_native)
cm_mod[cm_native == 0] = -1

print("Computation begins")

prob_it = np.zeros(cm_native.shape)
prob_id = np.zeros(cm_native.shape)

counter_it = 0
counter_id = 0
assert len(Q_sub) == len(GE)
assert len(Q_sub) == len(universes)
for q, ge, u in zip(Q_sub, GE, universes):
    assert len(q) == len(ge)
    assert len(q) == len(mask_fancy)
    for qt, get, mf  in zip(q, ge, mask_fancy):

        if qt > QlimIT[0] and qt < QlimIT[1] \
            and get > linear_it_low(qt) and get < linear_it_high(qt):

            ca_pos = u.trajectory[mf].positions
            self_distances = distances.self_distance_array(ca_pos)
            sq_dist_arr = np.zeros(cm_native.shape)
            triu = np.triu_indices_from(sq_dist_arr, k=1)
            sq_dist_arr[triu] = self_distances

            prob_it += (sq_dist_arr <= gamma*cm_mod)
            counter_it += 1

        elif qt > QlimIE[0] and qt < QlimIE[1] \
            and get > linear_ie_low(qt) and get < linear_ie_high(qt):

            ca_pos = u.trajectory[mf].positions
            self_distances = distances.self_distance_array(ca_pos)
            sq_dist_arr = np.zeros(cm_native.shape)
            triu = np.triu_indices_from(sq_dist_arr, k=1)
            sq_dist_arr[triu] = self_distances

            prob_id += (sq_dist_arr <= gamma*cm_mod)
            counter_id += 1

# normalization
prob_it /= counter_it
prob_id /= counter_id
print("Configurations used for IT: ", counter_it)
print("Configurations used for IE-: ", counter_id)
np.save(PROT_PATH / f"probability_IT_{dataset}.npy", prob_it)
np.save(PROT_PATH / f"probability_IE-_{dataset}.npy", prob_id)

# CM plots
fig, ax = plt.subplots(1, figsize=(10, 10))

#setup for plot
res_n = cm_native.shape[0]
cmap = sns.color_palette('mako_r', as_cmap=True)
# this are needed to center ticks to the "pixel"
centers = [1,res_n,res_n,1]
dx = (res_n - 1)/(res_n - 1)
extent = [centers[0]-dx/2, centers[1]+dx/2, centers[2]+dx/2, centers[3]-dx/2]

imint = ax.imshow(
    prob_it + prob_id.T,
    extent=extent,
    cmap=cmap)
ax_divider = make_axes_locatable(ax)
cax = ax_divider.append_axes("right", size="3%", pad=0.2)
cbar = fig.colorbar(imint, cax=cax)
cbar.set_label('Probability', labelpad=25)

# label and tiks
ax.set_xlabel('Residues')
ax.set_ylabel('Residues')
ax.xaxis.set_major_locator(plt.MultipleLocator(10))
ax.yaxis.set_major_locator(plt.MultipleLocator(10))
ax.grid(alpha=.5)

# info texts
textBox(x=0.93, y=0.945, s='IT', ax=ax)
textBox(x=0.02, y=0.03, s=r'$\mathrm{IE}_{-}$', ax=ax)

fig.savefig(PROT_PATH / 'img' / f'IE-_IT_cm_{dataset}.pdf', bbox_inches='tight')
