'''Compute and display the average of the local Q timeseries for 1ucs
These are used to understand why a folding traj fall in the kinetic intermediate'''
import numpy as np
import matplotlib
import matplotlib.pyplot as plt

from utils.paths import ANALYSIS_PATH

# Variables
prot_name = '1ucs'
dataset = '29Gen22_CV'
gamma = 1.2
ANALYSIS_PATH = ANALYSIS_PATH / prot_name

# graphical settings
matplotlib.rcParams['font.size'] = 24
matplotlib.rcParams['text.usetex'] = True
matplotlib.rcParams['font.family'] = 'modern computer'

# load timeseries
Qge = np.load(ANALYSIS_PATH / f'Q_local_thr0.4_{dataset}.npy') # contacts with |G'| > 0.4
QN = np.load(ANALYSIS_PATH / f'Q_local_RESrow0-6_REScol32-40_{dataset}.npy') # contacts related to the N terminus

# load folded, THR. and BT trajectories
folded = np.load(ANALYSIS_PATH / f'folded_trajectories_{dataset}.npy')
THR = np.load(ANALYSIS_PATH / f'THR_traj_{dataset}.npy')
BT = np.load(ANALYSIS_PATH / f'BT_traj_{dataset}.npy')
folded_filtered = [x for x in folded if x not in BT]
folded_filtered = [x for x in folded_filtered if x not in THR]

# Perform average for each trajectory
trajdim = Qge[0].shape
# NOTE: by folded we mean all the folded trajectories EXCLUDED BT and THR
Fg, c_Fg = np.zeros(trajdim), 0 # Folded; G' contacts
Ft, c_Ft = np.zeros(trajdim), 0 # Folded; threading contacts
Bg, c_Bg = np.zeros(trajdim), 0 # BT; G' contacts
Bt, c_Bt = np.zeros(trajdim), 0 # BT; threading contacts
Tg, c_Tg = np.zeros(trajdim), 0 # THR; G' contacts
Tt, c_Tt = np.zeros(trajdim), 0 # THR; threading contacts
for idx, (qge, qn) in enumerate(zip(Qge, QN)):
    if idx in folded_filtered:
        Fg += qge
        c_Fg += 1
        Ft += qn
        c_Ft += 1
    elif idx in BT:
        Bg += qge
        c_Bg += 1
        Bt += qn
        c_Bt += 1
    elif idx in THR:
        Tg += qge
        c_Tg += 1
        Tt += qn
        c_Tt += 1

# normalization
Fg /= c_Fg
Ft /= c_Ft
Bg /= c_Bg
Bt /= c_Bt
Tg /= c_Tg
Tt /= c_Tt

# ===================== plot
fig, axs = plt.subplots(3, 2, figsize=(20, 10))

# grid
for ax in axs.flat:
    ax.grid()

# folded
axs[0,0].plot(Fg)
axs[0,1].plot(Ft)

# BT
axs[1,0].plot(Bg)
axs[1,1].plot(Bt)

# THR
axs[2,0].plot(Tg)
axs[2,1].plot(Tt)

# labels
thr=0.4
betaRes = (0, 6) # N termionus
secAlphaRes = (32, 40) # second alpha helix and proximity
axs[0,0].set_title(f"$Q \; |G'| > {thr}$", pad=15)
axs[0,1].set_title(f"$Q$ Res:[{betaRes[0]:d}:{betaRes[1]:d}]-[{secAlphaRes[0]:d}:{secAlphaRes[1]:d}]", pad=15)

axs[0,0].set_ylabel('Folded directly')
axs[1,0].set_ylabel('BT')
axs[2,0].set_ylabel('THR')

fig.tight_layout()
fig.savefig(ANALYSIS_PATH / 'img' / f'Q_local_avg_{dataset}.pdf')