"""
Set of function used to generate the 2D histograms in the
2D Q vs GE plots. See Q_GE.ipynb notebook for their usage
"""
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import patches
from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

def hist2d(
    fig: matplotlib.figure.Figure,
    ax: plt.Axes,
    x: np.ndarray,
    y: np.ndarray,
    total_contacts: int,
    bins_y: int,
    scale: str,
    cmap: str,
    bar_pos: str,
    bar_label: str,
    pad_bar_label: int,
    xlabel: str,
    ylabel: str,
    title: str = None
    ) -> None:
    '''Plot bi-variate histograms with a colorbar

    NOTE: if the log option is selected, then all entries that are 0
    are considered = 1 to avoid graphical aberrations'''

    # Bin edges for Q
    bins_Q = [-1e-8, 1e-8] # this serves to catch 0 values
    for k in range(1, total_contacts-1):
        mid = (k/total_contacts + (k+1)/total_contacts)*0.5
        bins_Q.append(mid)
    bins_Q.append(1.0-1e-8) # this
    bins_Q.append(1.0+1e-8) # and this serves to catch ones

    h, xedges, yedges = np.histogram2d(
            x, y, 
            bins=[bins_Q, bins_y])
    h = h.T # to ensure x is the actual Cartesian def of x
    extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]

    aspect = 'auto'
    interpolation='gaussian'
    if scale == 'log':
        h[h == 0] = 1
        im = ax.imshow(
            h,
            cmap=cmap,
            extent=extent,
            aspect=aspect, # do NOT force pixels to be squares
            origin='lower',
            interpolation=interpolation,
            norm=matplotlib.colors.LogNorm())

    else:
        im = ax.imshow(
            h,
            cmap=cmap,
            extent=extent,
            aspect=aspect, # do NOT force pixels to be squares
            interpolation=interpolation,
            origin='lower')

    # colorbar setting
    if bar_pos == 'outside':
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.2)
        c = fig.colorbar(im, ax=ax, cax=cax)
        c.set_label(bar_label, rotation=90, labelpad=pad_bar_label)
    if bar_pos == 'inside':
        cbaxes = inset_axes(ax, height="3%", width="35%", 
                        loc='upper left', borderpad=1.7)
        c = fig.colorbar(im, ax=ax, cax=cbaxes, orientation='horizontal')
        c.ax.tick_params(labelsize=18)
        c.set_label(bar_label, labelpad=-pad_bar_label, fontsize=20)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title, pad=10, fontsize=25)

def free_energy2d(
    fig: matplotlib.figure.Figure,
    ax: plt.Axes,
    x: np.ndarray,
    y: np.ndarray,
    total_contacts: int,
    bins_y: int,
    cmap: str,
    bar_label: str,
    pad_bar_label: int,
    xlabel: str,
    ylabel: str,
    temp: float
    ) -> None:
    '''Plot bi-variate histograms with a colorbar

    NOTE: if the log option is selected, then all entries that are 0
    are considered = 1 to avoid graphical aberrations'''

    # Bin edges for Q
    bins_Q = [-1e-8, 1e-8] # this serves to catch 0 values
    for k in range(1, total_contacts-1):
        mid = (k/total_contacts + (k+1)/total_contacts)*0.5
        bins_Q.append(mid)
    bins_Q.append(1.0-1e-8) # this
    bins_Q.append(1.0+1e-8) # and this serves to catch ones

    h, xedges, yedges = np.histogram2d(
            x, y, 
            bins=[bins_Q, bins_y])
    h = h.T # to ensure x is the actual Cartesian def of x
    # Generate free energy
    logh = h.copy()
    logh[logh!=0] = -np.log(logh[logh!=0])/temp
    logh = logh + abs(logh.min())

    extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
    aspect = 'auto'
    interpolation='none'
    im = ax.imshow(
        logh,
        cmap=cmap,
        extent=extent,
        aspect=aspect, # do NOT force pixels to be squares
        origin='lower',
        interpolation=interpolation)

    # colorbar setting
    cbaxes = inset_axes(ax, height="3%", width="35%", 
                    loc='upper left', borderpad=1.7)
    c = fig.colorbar(im, ax=ax, cax=cbaxes, orientation='horizontal', ticks=plt.MultipleLocator(1))
    c.ax.tick_params(labelsize=18)
    c.set_label(bar_label, labelpad=-pad_bar_label, fontsize=18)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

def contour(
    fig: matplotlib.figure.Figure,
    ax: plt.Axes,
    x: np.ndarray,
    y: np.ndarray,
    total_contacts: int,
    bins_y: int,
    cmap: str,
    bar_label: str,
    pad_bar_label: int,
    xlabel: str,
    ylabel: str,
    temp: float
    ) -> None:
    '''Plot bi-variate histograms with a colorbar
    
    NOTE: if the log option is selected, then all entries that are 0
    are considered = 1 to avoid graphical aberrations'''

    # Bin edges for Q
    bins_Q = [-1e-8, 1e-8] # this serves to catch 0 values
    for k in range(1, total_contacts-1):
        mid = (k/total_contacts + (k+1)/total_contacts)*0.5
        bins_Q.append(mid)
    bins_Q.append(1.0-1e-8) # this
    bins_Q.append(1.0+1e-8) # and this serves to catch ones

    h, xedges, yedges = np.histogram2d(
            x, y, 
            bins=[bins_Q, bins_y])
    h = h.T # to ensure x is the actual Cartesian def of x
    # Generate free energy
    logh = h.copy()
    logh[logh!=0] = -np.log(logh[logh!=0])/temp
    logh = logh + abs(logh.min())

    extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
    aspect = 'auto'
    interpolation='none'
    im = ax.imshow(
        logh,
        cmap=cmap,
        extent=extent,
        aspect=aspect, # do NOT force pixels to be squares
        origin='lower',
        interpolation=interpolation)

    # colorbar setting
    cbaxes = inset_axes(ax, height="3%", width="35%", 
                    loc='upper left', borderpad=1.7)
    c = fig.colorbar(im, ax=ax, cax=cbaxes, orientation='horizontal', ticks=plt.MultipleLocator(1))
    c.ax.tick_params(labelsize=18)
    c.set_label(bar_label, labelpad=-pad_bar_label, fontsize=18)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

def arrow(
    ax: matplotlib.axes.Axes,
    xinit: float,
    yinit: float,
    xfin: float,
    yfin: float,
    rad: float,
    color: str
    ) -> None:
    style = "Simple, tail_width=0.5, head_width=7, head_length=8"
    kw_fol = dict(arrowstyle=style, color=color)
    arr = patches.FancyArrowPatch(
        (xinit, yinit), (xfin, yfin),
        alpha=.8,
        connectionstyle=f"arc3,rad={rad}",
        **kw_fol)
    ax.add_patch(arr)
