'''Miscellaneous collection of useful function used across multiple .py and .ipynb'''
from pathlib import Path
from typing import Tuple

import numpy as np

def get_run_id(simulation_path):
    """
    Extract the ID from the folder containing the MD run.
    The function assumes that the folder is named, for a protein
    with PDB code NXYZ, and run code ID as:
        Data_NXYZ_ID
    Hence, it returns ID as int.
    It handles the equilibrium simulations as follow:
    if the temperature is "1_050" then the int return is 1050

    Parameters
    ---------
    simulation_path : str or pathlib.Path
        simulation path to the directory containing the MD run

    Return
    -------
    index : int
        run ID
    """
    simulation_path = str(simulation_path)

    # extract index from idx
    data_dir_name = simulation_path.rsplit("/", maxsplit=1)[-1]
    assert data_dir_name.startswith("Data")

    if len(data_dir_name) == 11:
        # only one digit
        index = int(data_dir_name[-1])
    elif len(data_dir_name) == 12:
        # two digits
        index = int(data_dir_name[-2:])
    elif len(data_dir_name) == 13:
        # three digits
        index = int(data_dir_name[-3:])
    elif len(data_dir_name) == 15:
        # equilibrium run, where ID is a temperature like 1_050
        # In this case the index is 1_050 -> 1050
        index = int(data_dir_name[-5:].replace("_", ""))
    else:
        raise RuntimeError("Error in the extraction of the run ID from Data_protname_ID\n\
            Probably you are trying to use this function for a run described by its temperature")

    return index

def write_trajectory_info(path, origin_file, datasets_name, runs_list):
    """
    Write on file information regarding how the Q and RMSD trajectories
    are generated, in order to keep track of which file generates those
    and the mapping between runs (identified by an ID) and the timeseries

    The function saves:
    - the file path that generate the .npy
    - the map between run ID (unique) and the array index
    The structure is the one of a csv, with the first column having the
    array index and the second (comma-separated) the run ID.

    Hence, to recover which run is which one reads, e.g., the Q array
    and the file generated by this function. Then, Q[idx_first_column]
    correspond to run idx_second_column.

    The file can be loaded as a 2D Numpy array with the command:
    np.loadtxt(path/to/file, delimiter=',', dtype=int)

    Parameters
    ---------
    path : str or pathlib.Path
        Path where to save the csv file
    origin_file : str
        Name of the file from which this function as been called.
        This is used to keep track of where the file as been created.
    datasets_name : str
        str containing the names from which the dataset is taken.
        E.g.: 2Nov22_5Dic22_CV
    runs_list : List[str]
        List with the absolute paths to the Data directory where the
        actual trajectory is stored. This is necessary to call the
        get_run_id() function and extract the run ID
    """
    if not isinstance(path, Path):
        path = Path(path)
    if datasets_name.startswith("_"):
        datasets_name = datasets_name[1:]
    with open(path / f"ID_mapping_{datasets_name}.csv", "w", encoding="utf-8") as out:
        out.write(f"# Origin file: {origin_file}\n")
        out.write(f"# Dataset used: {datasets_name.split('_')}\n")
        out.write("# idx_array, run_id\n")

        for idx, run in enumerate(runs_list):
            out.write(f"{idx:d}, {get_run_id(run):d}\n")

def timeseries_mask(
    r: int,
    t: np.ndarray
    ) -> np.ndarray:
    '''Return a mask applicable to timeseries t;
    The first and the last point in the timeseries are selected

    Please note that if you want to use mask for fancy indexing
    you need to use np.nonzero(mask) instead'''

    assert t.ndim == 1

    mask = np.zeros(t.size, dtype=int)
    block = mask.size//r
    for i in range(r):
        mask[i*block] = 1
    mask[-1] = 1 # include last frame

    return mask

def block_avg(
    timeseries: np.ndarray,
    block_size: int
    ) -> Tuple[np.ndarray, np.ndarray]:
    '''Perfore block average over timeseries'''
    
    t = np.array(timeseries)
    N = t.shape[0]
    
    block_size = int(block_size)
    n_block = int(N/block_size) # number of blocks in the timeseries, round toward 0
    
    mean = np.zeros(n_block, dtype=np.float64)
    var = np.zeros(n_block, dtype=np.float64)
    for i in range(1, n_block+1):
        
        istr = (i-1)*block_size
        iend = i*block_size
        t_slice = t[istr:iend]
        
        mean[i-1] = t_slice.mean()
        var[i-1] = t_slice.var()
        
    return (mean, var)

def idx_plateau(
    timeseries: np.ndarray, 
    thr: float,
    len_plateau: int, 
    max_var: float
    ) -> int:
    '''Return index corresponding to the beginning 
    of the *first* plateau identified in timeseries
    
    Args:
        timeseries (array): timeseries where to search for a plateau
        thr (float): threshold above where the function search for the plateau
        len_plateau (int): len of the plateau; used to avoid 'local plateau'
        max_var (float): max variation allowed to identify a plateau
        
    The function search where the timeseries is greater then the thr arg and
    perform an average over the next len_plateau elements of the timeseries,
    if this falls in the range [value-max_var, value+max_var] then the index
    of the first element to be bigger then thr is considered the start fo the plateau
    '''
    
    indeces = np.where(timeseries > thr)[0]
    if indeces.size == 0:
        return None
    
    for idx in indeces:
        possible_plt = timeseries[idx: idx+len_plateau].mean()
        val = timeseries[idx]
        if (possible_plt > (val - max_var)) and (possible_plt < (val + max_var)):
            return idx
    
    return None