"""
Collection of useful functions to generate beautiful plots
"""
from typing import Dict, List

import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
# from mpl_toolkits.mplot3d.axes3d import Axes3D

def textBox(
    ax: plt.Axes,
    x: float,
    y: float,
    s: str
    ) -> None:
    ax.text(
        x=x, y=y,
        s=s,
        transform=ax.transAxes,
        bbox=dict(
            boxstyle="square",
            facecolor=(1,1,1)
            )
        )

def set_tickers(
    ax: plt.Axes,
    size_maj: float = 10,
    size_min: float = 7,
    width: float = 1.5,
    set_tick_pos: bool = False,
    x_maj: float = None,
    x_min: float = None,
    y_maj: float = None, 
    y_min: float = None
    ) -> None:
    '''Modifies the passed axes to change its style and obtain
    a beautiful scientific plot

    Please, use `fig, ax = plt.subplots()` to generate figures.
    '''
    ax.xaxis.set_tick_params(which='major', size=size_maj, width=width, direction='in', top='on')
    ax.xaxis.set_tick_params(which='minor', size=size_min, width=width, direction='in', top='on')
    ax.yaxis.set_tick_params(which='major', size=size_maj, width=width, direction='in', right='on')
    ax.yaxis.set_tick_params(which='minor', size=size_min, width=width, direction='in', right='on')

    if set_tick_pos:
        ax.xaxis.set_major_locator(mpl.ticker.MultipleLocator(x_maj))
        ax.xaxis.set_minor_locator(mpl.ticker.MultipleLocator(x_min))
        ax.yaxis.set_major_locator(mpl.ticker.MultipleLocator(y_maj))
        ax.yaxis.set_minor_locator(mpl.ticker.MultipleLocator(y_min))

def set_minor_grid(
    ax: plt.Axes,
    width: float = .25
    ) -> None:
    '''Set the minor grid on ax. This works best with:
    plt.style.use('seaborn-whitegrid')
    '''
    ax.xaxis.grid(True, "minor", linewidth=width)
    ax.yaxis.grid(True, "minor", linewidth=width)

def err_scatter(
    ax: plt.Axes,
    x: List[float],
    y: List[float],
    yerr: List[float],
    xlabel: str,
    ylabel: str
    ) -> None:
    '''Scatter plot with error bars on the y axis'''
    # sort arrays
    x_sorted, y_sorted, y_err_sorted = zip(*sorted(zip(x, y, yerr)))

    ax.errorbar(x_sorted, y_sorted, yerr=y_err_sorted, 
                linestyle='none', 
                fmt='.', c='deepskyblue', markersize=7,
                capsize=3, ecolor='black')

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

def imshow_colorbar(
    fig: mpl.figure.Figure,
    ax: plt.Axes,
    array: np.ndarray,
    cmap: str,
    bar_label: str,
    xlabel: str,
    ylabel: str,
    title: str = None
    ) -> None:
    '''plt.imshow() with a fixed colorbar'''

    im = ax.imshow(array, cmap=cmap)
    # colorbar setting
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    c = fig.colorbar(im, ax=ax, cax=cax)
    c.set_label(bar_label, rotation=90, labelpad=10)
    c.formatter.set_powerlimits((0, 0))
    c.update_ticks()

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title, pad=10)

def hist2d_colorbar(
    fig: mpl.figure.Figure,
    ax: plt.Axes,
    x: List[float],
    y: List[float],
    dict_hist: Dict,
    cmap: str,
    bar_label: str,
    pad_bar_label: int,
    xlabel: str,
    ylabel: str,
    title: str = None
    ) -> None:
    '''Plot bivariate histograms with a colorbar'''

    tot_a = dict_hist['total_contacts'][0]
    tot_b = dict_hist['total_contacts'][1]

    # Bin edges
    bins_Qa = [-1e-8, 1e-8] # this serves to catch 0 values
    for k in range(1, tot_a-1):
        mid = (k/tot_a + (k+1)/tot_a)*0.5
        bins_Qa.append(mid)
    bins_Qa.append(1.0-1e-8) # this
    bins_Qa.append(1.0+1e-8) # and this serves to catch ones

    bins_Qb = [-1e-8, 1e-8] # this serves to catch 0 values
    for k in range(1, tot_b-1):
        mid = (k/tot_b + (k+1)/tot_b)*0.5
        bins_Qb.append(mid)
    bins_Qb.append(1.0-1e-8) # this
    bins_Qb.append(1.0+1e-8) # and this serves to catch ones

    if dict_hist['scale'] == 'norm':
        _, _, _, h = ax.hist2d(x, y, bins=[bins_Qa, bins_Qb],
                                weights=np.repeat(1/x.size, x.shape[0]),
                                cmap=cmap)
    elif dict_hist['scale'] == 'log':
        _, _, _, h = ax.hist2d(x, y, bins=[bins_Qa, bins_Qb],
                                norm=mpl.colors.LogNorm(),
                                cmap=cmap)
    else:
        _, _, _, h = ax.hist2d(x, y, bins=[bins_Qa, bins_Qb],
                                cmap=cmap)

    # colorbar setting
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    c = fig.colorbar(h, ax=ax, cax=cax)
    c.set_label(bar_label, rotation=90, labelpad=pad_bar_label)

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title, pad=10, fontsize=25)

def hist2d_colorbar_inside(
    fig: mpl.figure.Figure,
    ax: plt.Axes,
    x: List[float],
    y: List[float],
    dict_hist: Dict,
    cmap: str,
    bar_label: str,
    pad_bar_label: int,
    xlabel: str, 
    ylabel: str,
    title: str = None
    ) -> None:
    '''Plot bivariate histograms with a colorbar insede the axes'''

    tot_a = dict_hist['total_contacts'][0]
    tot_b = dict_hist['total_contacts'][1]

    # Bin edges
    bins_Qa = [-1e-8, 1e-8] # this serves to catch 0 values
    for k in range(1, tot_a-1):
        mid = (k/tot_a + (k+1)/tot_a)*0.5
        bins_Qa.append(mid)
    bins_Qa.append(1.0-1e-8) # this
    bins_Qa.append(1.0+1e-8) # and this serves to catch ones

    bins_Qb = [-1e-8, 1e-8] # this serves to catch 0 values
    for k in range(1, tot_b-1):
        mid = (k/tot_b + (k+1)/tot_b)*0.5
        bins_Qb.append(mid)
    bins_Qb.append(1.0-1e-8) # this
    bins_Qb.append(1.0+1e-8) # and this serves to catch ones

    if dict_hist['scale'] == 'norm':
        _, _, _, h = ax.hist2d(x, y, bins=[bins_Qa, bins_Qb], 
                                weights=np.repeat(1/x.size, x.shape[0]), 
                                cmap=cmap)
    elif dict_hist['scale'] == 'log':
        _, _, _, h = ax.hist2d(x, y, bins=[bins_Qa, bins_Qb],
                                norm=mpl.colors.LogNorm(), 
                                cmap=cmap)
    else:
        _, _, _, h = ax.hist2d(x, y, bins=[bins_Qa, bins_Qb],
                                cmap=cmap)

    # colorbar setting
    cbaxes = inset_axes(ax, height="2.5%", width="35%", 
                        loc='lower right', borderpad=1.7)
    c = fig.colorbar(h, ax=ax, cax=cbaxes, orientation='horizontal')

    c.set_label(bar_label, rotation=0, labelpad=pad_bar_label)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title, pad=10, fontsize=25)
